#ifndef DPP_H
#define DPP_H

/* TOOD: lazy */
#include <wifidefs.h>
#ifdef USE_LIBDPP
#include <dpputils.h>
#endif

#ifndef USE_LIBDPP

#define DPP_PUB_AF_ACTION		0x09
#define DPP_PUB_AF_ACTION_OUI_TYPE	0x1A
#define DPP_PUB_AF_GAS_INITIAL_REQ	0x0A
#define DPP_PUB_AF_GAS_INITIAL_RESP	0x0B


#define FRAME_IS_DPP_PUB_AF(frame) \
		((*(frame) == DPP_PUB_AF_ACTION) && \
		(memcmp((frame + 1), "\x50\x6F\x9A", 3) == 0) && \
		((*(frame + 4)) == DPP_PUB_AF_ACTION_OUI_TYPE))

#define FRAME_IS_DPP_GAS_FRAME(frame) \
		((*(frame) == DPP_PUB_AF_GAS_INITIAL_REQ) || \
		(*(frame) == DPP_PUB_AF_GAS_INITIAL_RESP))

enum dpp_attribute_id {
	DPP_ATTR_STATUS = 0x1000,
	DPP_ATTR_I_BOOTSTRAP_KEY_HASH = 0x1001,
	DPP_ATTR_R_BOOTSTRAP_KEY_HASH = 0x1002,
	DPP_ATTR_I_PROTOCOL_KEY = 0x1003,
	DPP_ATTR_WRAPPED_DATA = 0x1004,
	DPP_ATTR_I_NONCE = 0x1005,
	DPP_ATTR_I_CAPABILITIES = 0x1006,
	DPP_ATTR_R_NONCE = 0x1007,
	DPP_ATTR_R_CAPABILITIES = 0x1008,
	DPP_ATTR_R_PROTOCOL_KEY = 0x1009,
	DPP_ATTR_I_AUTH_TAG = 0x100A,
	DPP_ATTR_R_AUTH_TAG = 0x100B,
	DPP_ATTR_CONFIG_OBJ = 0x100C,
	DPP_ATTR_CONNECTOR = 0x100D,
	DPP_ATTR_CONFIG_ATTR_OBJ = 0x100E,
	DPP_ATTR_BOOTSTRAP_KEY = 0x100F,
	DPP_ATTR_OWN_NET_NK_HASH = 0x1011,
	DPP_ATTR_FINITE_CYCLIC_GROUP = 0x1012,
	DPP_ATTR_ENCRYPTED_KEY = 0x1013,
	DPP_ATTR_ENROLLEE_NONCE = 0x1014,
	DPP_ATTR_CODE_IDENTIFIER = 0x1015,
	DPP_ATTR_TRANSACTION_ID = 0x1016,
	DPP_ATTR_BOOTSTRAP_INFO = 0x1017,
	DPP_ATTR_CHANNEL = 0x1018,
	DPP_ATTR_PROTOCOL_VERSION = 0x1019,
	DPP_ATTR_ENVELOPED_DATA = 0x101A,
	DPP_ATTR_SEND_CONN_STATUS = 0x101B,
	DPP_ATTR_CONN_STATUS = 0x101C,
	DPP_ATTR_RECONFIG_FLAGS = 0x101D,
	DPP_ATTR_C_SIGN_KEY_HASH = 0x101E,
	DPP_ATTR_CSR_ATTR_REQ = 0x101F,
	DPP_ATTR_A_NONCE = 0x1020,
	DPP_ATTR_E_PRIME_ID = 0x1021,
	DPP_ATTR_CONFIGURATOR_NONCE = 0x1022,
};

enum dpp_public_action_frame_type {
	DPP_PA_AUTHENTICATION_REQ = 0,
	DPP_PA_AUTHENTICATION_RESP = 1,
	DPP_PA_AUTHENTICATION_CONF = 2,
	DPP_PA_PEER_DISCOVERY_REQ = 5,
	DPP_PA_PEER_DISCOVERY_RESP = 6,
	DPP_PA_PKEX_V1_EXCHANGE_REQ = 7,
	DPP_PA_PKEX_EXCHANGE_RESP = 8,
	DPP_PA_PKEX_COMMIT_REVEAL_REQ = 9,
	DPP_PA_PKEX_COMMIT_REVEAL_RESP = 10,
	DPP_PA_CONFIGURATION_RESULT = 11,
	DPP_PA_CONNECTION_STATUS_RESULT = 12,
	DPP_PA_PRESENCE_ANNOUNCEMENT = 13,
	DPP_PA_RECONFIG_ANNOUNCEMENT = 14,
	DPP_PA_RECONFIG_AUTH_REQ = 15,
	DPP_PA_RECONFIG_AUTH_RESP = 16,
	DPP_PA_RECONFIG_AUTH_CONF = 17,
	DPP_PA_PKEX_EXCHANGE_REQ = 18,
	DPP_PA_PB_PRESENCE_ANNOUNCEMENT = 19,
	DPP_PA_PB_PRESENCE_ANNOUNCEMENT_RESP = 20,
	DPP_PA_PRIV_PEER_INTRO_QUERY = 21,
	DPP_PA_PRIV_PEER_INTRO_NOTIFY = 22,
	DPP_PA_PRIV_PEER_INTRO_UPDATE = 23,
};

uint8_t *dpp_get_attr(uint8_t *msg, uint16_t msglen, uint16_t attr, uint16_t *olen);

uint8_t dpp_get_pub_af_type(const uint8_t *frame, uint16_t framelen);
uint8_t dpp_get_gas_frame_type(const uint8_t *frame, uint16_t framelen);
uint8_t dpp_get_frame_type(const uint8_t *frame, uint16_t framelen);

#endif

#define FRAME_DST_ENROLLEE(type) \
		((type == DPP_PA_AUTHENTICATION_REQ) || \
		 (type == DPP_PA_AUTHENTICATION_CONF))

#define DPP_KEY_FILE "/etc/multiap/dpp_keys.json"


enum dpp_responder_state {
	DPP_BOOSTRAPPING,
	DPP_AUTHENTICATING,
	DPP_AUTHENTICATED
};

/* taken from dpp.h */
enum dpp_bootstrap_type {
	DPP_BOOTSTRAP_QR_CODE_WIFI,
	DPP_BOOTSTRAP_PKEX,
	DPP_BOOTSTRAP_NFC_URI,
};

struct dpp_frame {
	uint8_t type;
	uint16_t len;
	uint8_t *data;
	struct timespec time;

	struct list_head list;
};

struct dpp_chirp {
	struct agent *a;
	//uint8_t hashlen;
	uint8_t hash[32]; /* derived from DPP frame(s) */

	struct dpp_frame *auth_req;

	struct uloop_fd chirp_fd;
	struct list_head list;
	struct atimer timeout;
};

enum dpp_onboarded {
	DPP_CONFOBJ_SSID  = 1 << 0,
	DPP_CONFOBJ_KEY  = 1 << 1,
	DPP_CONFOBJ_AKM  = 1 << 2,
	DPP_CONNECTOR  = 1 << 3,
	DPP_CSIGN = 1 << 4,
	DPP_NETACCESSKEY = 1 << 5
};

struct dpp_creds {
	uint16_t stored; /* TODO: rename */

	char ssid[33];
	char key[65];
	char akm[32];
	char *connector;
	char *csign;
	char *netaccesskey;
};

struct dpp_chan {
	uint8_t channel;
	uint8_t opclass;
	uint32_t freq;
};

#define DPP_MAX_CH_OPCLASS 16
/* TODO: this struct should be linked against a node? */
struct dpp_enrollee {
	struct agent *a;
	uint8_t mac[6];
	char uri[128];

	bool onboarded;
	uint8_t bootstrap_id;

	enum wifi_band band;
	int num_chan;
	struct dpp_chan chan[DPP_MAX_CH_OPCLASS];
	enum dpp_bootstrap_type type;

	struct dpp_chirp *chirp;

	struct list_head framelist;
	struct list_head list;
};

struct dpp_context_wifi {
	struct uloop_fd dpp_sk;
	int dpp_fd;

	int ec_len;
	uint8_t *ec_key;

	struct list_head chirplist;
	struct list_head enrolleelist;
};

struct netif_bk;
struct agent;
enum dpp_event;

char *agent_get_backhaul_hex(struct agent *a, enum wifi_band band,
			   char *encryption, char *hex_ssid, char *hex_key);

const char *dpp_type2str(enum dpp_bootstrap_type type);
const char *dpp_frame_type2str(uint8_t type);
const char *dpp_state2str(enum dpp_responder_state state);

uint8_t dpp_get_pub_af_type(const uint8_t *frame, uint16_t framelen);
uint8_t dpp_get_gas_frame_type(const uint8_t *frame, uint16_t framelen);
uint8_t dpp_get_frame_type(const uint8_t *frame, uint16_t framelen);

struct dpp_chirp *dpp_chirp_get(struct agent *a, uint8_t *hash);
struct dpp_chirp *dpp_chirp_alloc(struct agent *a, uint8_t *hash);


struct dpp_enrollee *dpp_enrollee_get(struct agent *a, uint8_t *mac);
struct dpp_enrollee *dpp_enrollee_alloc(struct agent *a, uint8_t *mac);
void dpp_enrollee_free(struct dpp_enrollee *e);

enum wifi_band wifi_channel_to_band(uint8_t channel);

/* TODO: how to combine with dpp_gen_uri()? */
int dpp_gen_uri_wifi(struct netif_bk *bk);
int dpp_del_uri(struct netif_bk *bk);
int dpp_chirp(struct netif_bk *bk);
int dpp_listen_enrollee(struct netif_bk *bk);
int dpp_parse_uri_chan_list(struct agent *a, struct dpp_enrollee *e, char *uri);
void dpp_sta_event_handler(void *c, struct blob_attr *msg);

int dpp_enrollee_update_hash(struct agent *a,
			     struct dpp_enrollee *e,
			     uint8_t len, uint8_t *hash);
struct dpp_frame *dpp_enrollee_get_frame(struct agent *a,
					 struct dpp_enrollee *e, uint8_t type);
struct dpp_frame *dpp_enrollee_add_frame(struct agent *a,
					 struct dpp_enrollee *e,
					 uint8_t type, uint16_t len,
					 uint8_t *data);


int dpp_get_keys(struct agent *a);
int dpp_gen_keypair(struct agent *a);

void dpp_period_chirp(atimer_t *t);

void dpp_configurator_evh(void *agent, struct blob_attr *msg);
void dpp_enrollee_evh(void *agent, struct blob_attr *msg);

uint16_t dpp_get_controller_tcp_port(struct agent *a, enum wifi_band band);

void dpp_event_uloop_cb(struct uloop_fd *fd, unsigned int events);

int dpp_init_context(struct agent *a);

int dpp_frame_handler(void *dpp, uint8_t *smac, enum dpp_event ev,
		      uint8_t *frame, size_t framelen);

void dpp_legacy_timeout_cb(atimer_t *t);
#endif
